
%% load admittance data, convert to impedance data & select frequency range

clc
clear all

impedancedata = readmatrix('PN_500mV.csv');

frequency = impedancedata(2:end,1); % in Hz
omega = 2*pi*frequency;
admittance_G = impedancedata(2:end,2); 
admittance_B = impedancedata(2:end,3); 
impedance_R = admittance_G./(admittance_G.^2+admittance_B.^2);
impedance_X = -admittance_B./(admittance_G.^2+admittance_B.^2);
impedance_Mag = sqrt(impedance_R.^2+impedance_X.^2);
impedance_phase = atand(impedance_X./impedance_R);

%select frequency range
f_min = 1*10^-5; 
f_max =  1*10^9; 

[valuemin,indexmin]=min(abs(f_min-frequency)); f_min = frequency(indexmin);
[valuemax,indexmax]=min(abs(f_max-frequency)); f_max = frequency(indexmax);
frequency = frequency(indexmin:indexmax); omega = omega(indexmin:indexmax);
impedance_Mag = impedance_Mag(indexmin:indexmax); impedance_phase = impedance_phase(indexmin:indexmax); impedance_R = impedance_R(indexmin:indexmax); impedance_X = impedance_X(indexmin:indexmax);

Rs_init = min(impedance_R)-0.001;

%% global CNLS fitting for initial values
% impedance model contains a PN junction, series resistance, and inductor

% set initial value, lower bound, and upper bound for all the fitting parameters 

% PN junction resistance
RD_init = impedance_Mag(1); 
RD_low = 0; %Ohm, lower bound
RD_high = Inf; %Ohm, higher bound

% PN junction capacitance 
C_init = 1; % in uF 
C_low = 0; % in uF, lower bound
C_high = Inf; % in uF, upper bound

% Series resistance
Rser = Rs_init*1000; %Rser_LOANA; %mOhm, fill in value from LOANA
Rser_low = 0; %mOhm, lower bound
Rser_high = Rser*1.2; %mOhm, upper bound

omegafit = [omega omega];
impedancefit = [impedance_R impedance_X];

fun1 = @(fit,omegafit) fit(3)+fit(1)./(1+(fit(1)^2)*(omegafit.^2).*(fit(2).^2));
fun2 = @(fit,omegafit) -((fit(1)^2)*omegafit.*fit(2))./(1+(fit(1)^2)*(omegafit.^2).*(fit(2).^2));

fun = @(fit,omegafit) [fun1(fit,omegafit(:,1)),  fun2(fit,omegafit(:,2))];

fit0 = [RD_init C_init*10^-6 Rser/1000];
lb = [RD_low C_low Rser_low/1000];
ub = [RD_high C_high Rser_high/1000];

fit = lsqcurvefit(fun,fit0,omegafit,impedancefit,lb,ub);
RD_global = fit(1); C_global = fit(2); 

%% second iteration CNLS fitting

omegafit = [omega omega];
impedancefit = [impedance_R impedance_X];

fun1 = @(fit,omegafit) fit(3)+fit(1)./(1+(fit(1)^2)*(omegafit.^2).*(fit(2).^2));
fun2 = @(fit,omegafit) -((fit(1)^2)*omegafit.*fit(2))./(1+(fit(1)^2)*(omegafit.^2).*(fit(2).^2));

fun = @(fit,omegafit) [fun1(fit,omegafit(:,1)),  fun2(fit,omegafit(:,2))];

fit0 = [RD_global C_global Rser/1000];
lb = [RD_low C_low Rser_low/1000];
ub = [RD_high C_high Rser_high/1000];

fit = lsqcurvefit(fun,fit0,omegafit,impedancefit,lb,ub);
Rj_fit = fit(1); Cj_fit = fit(2); Rs_fit = fit(3);

disp(['CNLS best-fit values are Cj = ',num2str(Cj_fit*10^6),' uF/cm^2, ', ...
     'Rj = ',num2str(Rj_fit),' Ohm*cm^2 ',...
     'Rs = ',num2str(Rs_fit*1000),' mOhm*cm^2 ']);


%% Determine global goodness of CNLS fit

Real_fit_full = fit(3)+fit(1)./(1+(fit(1)^2)*(omega.^2).*(fit(2).^2));
Compl_fit_full = -((fit(1)^2)*omega.*fit(2))./(1+(fit(1)^2)*(omega.^2).*(fit(2).^2));
Mag_fit_full = sqrt(Real_fit_full.^2+Compl_fit_full.^2);
Angle_fit_full = atand(Compl_fit_full./Real_fit_full);

%calculate NRMSE magnitude
MSE_mag = goodnessOfFit(impedance_Mag,Mag_fit_full,'MSE');
RMSE_mag = sqrt(MSE_mag);
NRMSE_mag = abs(RMSE_mag/(max(impedance_Mag)-min(impedance_Mag)));

%calculate RMSE phase
MSE_phase = goodnessOfFit(impedance_phase,Angle_fit_full,'MSE');
RMSE_phase = sqrt(MSE_phase);
NRMSE_phase = abs(RMSE_phase/(max(impedance_phase)-min(impedance_phase)));

%calculate NRMSE resistance
MSE_real = goodnessOfFit(impedance_R,Real_fit_full,'MSE');
RMSE_real = sqrt(MSE_real);
NRMSE_real = abs(RMSE_real/(max(impedance_R)-min(impedance_R)));

%calculate NRMSE reactance
MSE_complex = goodnessOfFit(impedance_X,Compl_fit_full,'MSE');
RMSE_complex = sqrt(MSE_complex);
NRMSE_complex = abs(RMSE_complex/(max(impedance_X)-min(impedance_X)));

%% Analytic method

for i=1:length(omega)

R = impedance_R(i);
X = impedance_X(i);
om = omega(i);
V1 = Rs_init;

syms V2 V3

eqn1 = (V1 + (V2/(1+(V2*om*V3)^2))) == R;
eqn2 = (-((V2^2)*om*V3)/(1+(V2*om*V3)^2)) == X;

eqns = [eqn1,eqn2];
S = solve(eqns);

Rj_lin(i) = double(S.V2);
Cj_lin(i) = double(S.V3);
end

%% circuit element values from CNLS method
Rj_fit_rec = Rj_fit;
Rs_fit_rec = Rs_fit;
Cj_fit_rec = Cj_fit;

%% circuit element values from Analytic method as function of frequency

Rj_an_rec = transpose(Rj_lin);
Rs_an_rec = Rs_init; 
Cj_an_rec = transpose(Cj_lin);

%% determine low-frequency analytic values of RC loop elements

Rj_lowf = mean(Rj_an_rec(1:10));
Cj_lowf = mean(Cj_an_rec(1:10));

%% identify frequency where R and C drop (named f_C and f_R in article)

idx_Rj = find(Rj_an_rec<0.9*Rj_lowf,1);
idx_Cj = find(Cj_an_rec<0.9*Cj_lowf,1);

fmax_R = frequency(idx_Rj);
fmax_C = frequency(idx_Cj);

disp(['f_C = ',num2str(fmax_C),' Hz ',...
     'f_R = ',num2str(fmax_R),' Hz ']);

%% reconstruct impedance

%CNLS reconstruction
rec_R_fit = Rs_fit_rec + (Rj_fit_rec)./(1+(Rj_fit_rec.*omega.*Cj_fit_rec).^2);
rec_X_fit = -((Rj_fit_rec^2).*omega.*Cj_fit_rec)./(1+(Rj_fit_rec.*omega.*Cj_fit_rec).^2);

%Analytic reconstruction
rec_R_an = Rs_an_rec + (Rj_an_rec)./(1+(Rj_an_rec.*omega.*Cj_an_rec).^2);
rec_X_an = -((Rj_an_rec.^2).*omega.*Cj_an_rec)./(1+(Rj_an_rec.*omega.*Cj_an_rec).^2);


%% create plot

%plot style
plotstyle_data = '-';
plotstyle_fit1 = '-.';
plotstyle_fit2 = '--';
marker_size = 10;
marker_size_fit = 6;
line_width_data = 1;
line_width_fit = 2.5;
color_data = '#000000';
color_fit1 = '#4DBEEE'; 
color_fit2 = 'red'; 
freq_min = f_min;
freq_max = f_max;

close all
figure
tiledlayout(4,1)

nexttile
loglog(frequency,impedance_R,plotstyle_data,'MarkerSize',marker_size,'LineWidth',line_width_data,'Color',color_data)
hold on 
loglog(frequency,rec_R_fit,plotstyle_fit1,'MarkerSize',marker_size_fit,'LineWidth',line_width_fit,'Color',color_fit1)
loglog(frequency,rec_R_an,plotstyle_fit2,'MarkerSize',marker_size_fit,'LineWidth',line_width_fit,'Color',color_fit2)
grid on
legend('Data','CNLS','Analytic','Location','southwest')
set(gca,'FontSize',12)
xlabel('Frequency (Hz)','FontSize',14) 
ylabel('Z''(\Omega)','FontSize',14)
xlim([freq_min freq_max])
ylim([min(impedance_R)/2 max(impedance_R)*2])


nexttile
loglog(frequency,impedance_X,plotstyle_data,'MarkerSize',marker_size,'LineWidth',line_width_data,'Color',color_data)
hold on 
loglog(frequency,rec_X_fit,plotstyle_fit1,'MarkerSize',marker_size_fit,'LineWidth',line_width_fit,'Color',color_fit1)
loglog(frequency,rec_X_an,plotstyle_fit2,'MarkerSize',marker_size_fit,'LineWidth',line_width_fit,'Color',color_fit2)
grid on
legend('Data','CNLS','Analytic','Location','northwest')
set(gca,'FontSize',12)
xlabel('Frequency (Hz)','FontSize',14) 
ylabel('Z''''(\Omega)','FontSize',14)
xlim([freq_min freq_max])
ylim([min(impedance_X)*10 max(impedance_X)/10]) %values are negative

Cj_fit_plot = Cj_fit*ones(length(omega),1);
Rj_fit_plot = Rj_fit*ones(length(omega),1);

nexttile
semilogx(frequency,Cj_fit_plot*10^6,'LineWidth',line_width_fit,'Color',color_fit1)
hold on
semilogx(frequency,Cj_lin*10^6,'--','LineWidth',line_width_fit,'Color',color_fit2)
xline(fmax_C,':','LineWidth',line_width_fit)
grid on
set(gca,'FontSize',12)
xlabel('Frequency (Hz)','FontSize',14) 
ylabel('C_{j}(\muF)','FontSize',14)
legend('CNLS','Analytic','f_{C}','Location','southwest')
xlim([freq_min freq_max])
ylim([0 Cj_fit*1.2*10^6])

nexttile
semilogx(frequency,Rj_fit_plot,'LineWidth',line_width_fit,'Color',color_fit1)
hold on
semilogx(frequency,Rj_lin,'--','LineWidth',line_width_fit,'Color',color_fit2)
xline(fmax_R,':','LineWidth',line_width_fit)
grid on
set(gca,'FontSize',12)
xlabel('Frequency (Hz)','FontSize',14) 
ylabel('R_{j}(\Omega)','FontSize',14)
legend('CNLS','Analytic','f_{R}','Location','southwest')
ylim([0 Rj_fit*1.2])
xlim([freq_min freq_max])


set(gcf,'position',[50,0,550,1000])


